# -*- coding: utf-8 -*-

from __future__ import (absolute_import, division, print_function, unicode_literals)


class BaseBot(object):
    def __init__(self, channel_client=None, storage_client=None, nlu_client_factory=None, event=None):
        self.channel_client = channel_client
        self.storage_client = storage_client
        self.nlu_client_factory = nlu_client_factory
        self.event = event

    def handle_message(self, event, context):
        raise NotImplementedError()

    def send_message(self, message, user_id=None, channel=None, extra=None):
        self.channel_client.send_message(user_id, message, channel, event=self.event, extra=extra)

    def set_project_data(self, data):
        self.storage_client.set_project_data(data)

    def get_project_data(self):
        return self.storage_client.get_project_data()

    def set_user_data(self, data, user_id=None, channel=None):
        _user_id = user_id or self.event.get('sender', {}).get('id')
        _channel = channel or self.event.get('channel')
        self.storage_client.set_user_data(data, user_id=_user_id, channel=_channel)

    def get_user_data(self, user_id=None, channel=None):
        _user_id = user_id or self.event.get('sender', {}).get('id')
        _channel = channel or self.event.get('channel')
        return self.storage_client.get_user_data(user_id=_user_id, channel=_channel)

    def nlu(self, vendor):
        return self.nlu_client_factory.get(vendor)



# -*- coding: utf-8 -*-

from __future__ import (absolute_import, division, print_function, unicode_literals)

import traceback


def traceback_to_string(exc):
    s = traceback.extract_stack()[:-3] + traceback.extract_tb(exc.__traceback__)
    l = traceback.format_list(s)
    return ''.join(l) + '\\n  {} {}'.format(exc.__class__, exc)



# -*- coding: utf-8 -*-

from __future__ import (absolute_import, division, print_function, unicode_literals)

import json

from bothub_client.messages import Message
from bothub_client.transports import HttpTransport
from bothub_client.transports import ZmqTransport


def get_channel_client(context):
    scheme_to_channel_client = {
        'http': ChannelClient,
        'https': ChannelClient,
        'tcp': ZmqChannelClient,
    }

    endpoint = context.get('channel', {}).get('endpoint', 'http:')
    scheme = endpoint.split(':')[0]
    client_class = scheme_to_channel_client.get(scheme, ZmqChannelClient)
    return client_class.init_client(context)


def handle_message(event, context, bot_class):
    channel = get_channel_client(context)
    storage = StorageClient.init_client(context, event=event)
    nlu_client_factory = NluClientFactory(context)

    bot = bot_class(channel_client=channel, storage_client=storage, nlu_client_factory=nlu_client_factory, event=event)
    response = bot.handle_message(event, context)
    return {'response': response}


class Client(object):
    def __init__(self, project_id, api_key, base_url, transport=None):
        self.project_id = project_id
        self.base_url = base_url
        self.api_key = api_key
        self.transport = transport or HttpTransport(base_url)


class ChannelClient(Client):
    def __init__(self, project_id, api_key, base_url, transport=None):
        _transport = transport or HttpTransport
        super(ChannelClient, self).__init__(project_id, api_key, base_url, transport)

    @staticmethod
    def init_client(context):
        project_id = context.get('project_id')
        api_key = context.get('api_key', '')
        channel_endpoint = context.get('channel', {}).get('endpoint')
        return ChannelClient(project_id, api_key, channel_endpoint)

    def send_message(self, chat_id, message, channel=None, event=None, extra=None):
        sender_id = event.get('sender', {}).get('id', None)
        origin_channel = event.get('channel')
        _chat_id = chat_id or sender_id
        _channel = channel or origin_channel
        self.transport.post('/messages', {
            'channel': _channel,
            'receiver': _chat_id,
            'message': message,
            'event': event,
            'context': {
                'project_id': self.project_id,
                'api_key': self.api_key
            },
            'extra': extra
        })


class ZmqChannelClient(Client):
    def __init__(self, project_id, api_key, base_url, transport=None, context=None):
        _transport = transport or ZmqTransport(base_url)
        self.context = context
        super(ZmqChannelClient, self).__init__(project_id, api_key, base_url, _transport)

    @staticmethod
    def init_client(context):
        project_id = context.get('project_id')
        api_key = context.get('api_key', '')
        channel_endpoint = context.get('channel', {}).get('endpoint')
        return ZmqChannelClient(project_id, api_key, channel_endpoint, context=context)

    def get_channel_obj(self, channel_type):
        channels = self.context['channel']['channels']
        for c in channels:
            if c['type'] == channel_type:
                return c

    def send_message(self, chat_id, message, channel=None, event=None, extra=None):
        sender_id = event.get('sender', {}).get('id', None)
        origin_channel = event.get('channel')
        _chat_id = chat_id or sender_id
        channel_type = channel or origin_channel
        _channel = self.get_channel_obj(channel_type)
        data = {
            'channel': _channel,
            'receiver': _chat_id,
            'event': event,
            'context': {
                'project_id': self.project_id,
                'api_key': self.api_key
            },
            'extra': extra
        }

        if isinstance(message, Message):
            data['message'] = {
                'model': message.model,
                'event': message.event
            }
        else:
            data['message'] = message

        self.transport.send_multipart([json.dumps(data).encode('utf8')])


class StorageClient(Client):
    def __init__(self, project_id, api_key, base_url, transport=None, user=None):
        super(StorageClient, self).__init__(project_id, api_key, base_url, transport)
        self.current_user = user

    @staticmethod
    def init_client(context, event=None, transport=None):
        project_id = context.get('project_id')
        api_key = context.get('api_key', '')
        storage_endpoint = context.get('storage', {}).get('endpoint')
        user = None if not event else (event['channel'], event['sender']['id'])
        return StorageClient(project_id, api_key, storage_endpoint, transport=transport, user=user)

    def set_project_data(self, data):
        self.transport.put(
            '/projects/{}'.format(self.project_id),
            {'data': data},
        )

    def get_project_data(self):
        return self.transport.get(
            '/projects/{}'.format(self.project_id),
        ).get('data') or {}

    def set_user_data(self, channel, user_id, data):
        self.transport.put(
            '/projects/{}/channels/{}/users/{}'.format(self.project_id, channel, user_id),
            {'data': data},
        )

    def get_user_data(self, channel, user_id):
        return self.transport.get(
            '/projects/{}/channels/{}/users/{}'.format(self.project_id, channel, user_id)
        ).get('data') or {}

    def set_current_user_data(self, data):
        channel, user_id = self.current_user
        self.set_user_data(channel, user_id, data)

    def get_current_user_data(self):
        channel, user_id = self.current_user
        return self.get_user_data(channel, user_id)


class NluClient(object):
    ''''''
    def ask(self, event=None, message=None, session_id=None):
        '''Request a query to NLU service.

        There are two styles to request to NLU service::

        * client.ask(event=event)
        * client.ask(message=message, session_id=session_id)
        '''
        raise NotImplementedError()


class ApiAiNluClient(NluClient):
    def __init__(self, api_key):
        self.api_key = api_key
        apiai_module = __import__('apiai')
        self.apiai = apiai_module.ApiAI(api_key)

    def parse_response(self, response):
        content = response.read().decode('utf-8')
        json_response = json.loads(content)
        action = NluAction(
            json_response.get('result').get('action'),
            json_response.get('result').get('parameters'),
            json_response.get('result').get('actionIncomplete')
        )
        next_message = json_response.get('result').get('fulfillment', {}).get('speech')
        return NluResponse(json_response, next_message, action)

    def ask(self, event=None, message=None, session_id=None):
        if event:
            return self.ask_with_event(event)
        if message and session_id:
            return self.ask_with_message(message, session_id)

    def ask_with_event(self, event):
        request = self.apiai.text_request()
        request.query = event.get('content')
        request.session_id = '{}-{}'.format(event.get('channel'), event.get('sender').get('id'))
        response = request.getresponse()
        return self.parse_response(response)

    def ask_with_message(self, message, session_id):
        request = self.apiai.text_request()
        request.query = message
        request.session_id = session_id
        response = request.getresponse()
        return self.parse_response(response)


class NluClientFactory(object):
    NAME_TO_CLIENT = {
        'apiai': ApiAiNluClient
    }

    def __init__(self, context):
        self.integrations_params = context.get('nlu')

    def get(self, vendor):
        '''Returns a proper NluClient object'''
        return self.NAME_TO_CLIENT.get(vendor)(**self.integrations_params.get(vendor))


class NluAction(object):
    '''A NluAction class represents an intent identified action.'''

    def __init__(self, intent=None, parameters=None, action_incomplete=True):
        self.intent = intent
        self.parameters = parameters or {}
        self.completed = not action_incomplete

    def __repr__(self):
        return "<NluAction '{}', completed: {}, args: {}>".format(
            self.intent,
            self.completed,
            self.parameters
        )


class NluResponse(object):
    '''A NluResponse class containing a response from a NLU service.'''
    def __init__(self, response, next_message=None, action=None):
        self.raw_response = response
        self.action = action
        self.next_message = next_message



# -*- coding: utf-8 -*-

from __future__ import (absolute_import, division, print_function, unicode_literals)

import requests
import zmq


class HttpTransport(object):
    def __init__(self, base_url=''):
        self.session = requests.Session()
        self.base_url = base_url

    def get(self, path):
        return self.session.get('{}{}'.format(self.base_url, path)).json()

    def post(self, path, data=None):
        return self.session.post('{}{}'.format(self.base_url, path), json=data)

    def put(self, path, data=None):
        return self.session.put('{}{}'.format(self.base_url, path), json=data)


class ZmqTransport(object):
    def __init__(self, address):
        self.context = zmq.Context()
        self.socket = self.context.socket(zmq.PUSH)
        self.socket.connect(address)

    def send_json(self, data):
        return self.socket.send_json(data)

    def send_multipart(self, data):
        return self.socket.send_multipart(data)



# -*- coding: utf-8 -*-

from __future__ import (absolute_import, division, print_function, unicode_literals)


class MockTransport(object):
    def __init__(self):
        self.recorded = []

    def record(self, data):
        self.recorded.append(data)

    def put(self, url, data):
        return self.recorded.pop(0) if self.recorded else None

    def get(self, url, data):
        return self.recorded.pop(0) if self.recorded else None



# -*- coding: utf-8 -*-


class Message(object):
    def __init__(self, event):
        self.model = []
        self.event = event

    def set_text(self, text):
        self.model.append({
            'command': 'set_text',
            'args': {
                'text': text
            }
        })
        return self

    def add_url_button(self, text, url):
        self.model.append({
            'command': 'add_url_button',
            'args': {
                'text': text,
                'url': url
            }
        })
        return self

    def add_postback_button(self, text, payload):
        self.model.append({
            'command': 'add_postback_button',
            'args': {
                'text': text,
                'payload': payload
            }
        })
        return self

    def add_quick_reply(self, text, image_url=None, payload=None):
        self.model.append({
            'command': 'add_quick_reply',
            'args': {
                'text': text,
                'image_url': image_url,
                'payload': payload,
            }
        })
        return self

    def add_location_request(self, text):
        self.model.append({
            'command': 'add_location_request',
            'args': {
                'text': text
            }
        })
        return self

    def add_keyboard_button(self, text):
        self.model.append({
            'command': 'add_keyboard_button',
            'args': {
                'text': text
            }
        })
        return self

    def __repr__(self):
        s = []
        for m in self.model:
            s.append('{} {}'.format(m['command'], m['args']))
        return '\n'.join(s)



# -*- coding: utf-8 -*-

from __future__ import (absolute_import, division, print_function, unicode_literals)

import requests_mock

from bothub_client.bot import BaseBot
from bothub_client.clients import handle_message


class Bot(BaseBot):
    def handle_message(self, event, context):
        content = event.get('content')
        self.send_message('hello, you said: {}'.format(content))


def test_handle_message_should_add_channel_cmd_buff():
    event = {
        'content': 'hi!',
        'channel': 'mychannel',
        'sender': {
            'id': 'abcd1234',
            'name': 'myname'
        }
    }

    context = {
        'channel': {
            'endpoint': 'http://localhost'
        }
    }

    with requests_mock.mock() as m:
        m.post('http://localhost/')
        response = handle_message(event, context, Bot)
        assert response['response'] is None
        assert m.called
        assert m.request_history[0].json() == {
            'channel': 'mychannel',
            'receiver': 'abcd1234',
            'message': 'hello, you said: hi!',
            'event': event
        }



# -*- coding: utf-8 -*-

from __future__ import (absolute_import, division, print_function, unicode_literals)

import requests_mock

from bothub_client.clients import StorageClient
from .testutils import MockTransport


def fixture_client(transport=None):
    context = {
        'project_id': 1,
        'api_key': 'testkey',
        'storage': {
            'endpoint': 'http://192.168.0.1:9000'
        }
    }

    client = StorageClient.init_client(context, transport=transport)
    return client


def record_project_data(transport):
    transport.record({'score': 11})


def test_init_client_should_returns_client_obj():
    client = fixture_client()
    assert client.project_id == 1
    assert client.api_key == 'testkey'
    assert client.base_url == 'http://192.168.0.1:9000'


def test_set_project_data():
    transport = MockTransport()

    client = fixture_client(transport)
    client.set_project_data({
        'score': 11
    })
